c
c   shifted obj(nx,x)
c
        double precision function sh_obj(n,x)
        double precision x(*)

        parameter(nx=2)

        logical hasdiscrete, needshift
        common /hasdiscrete/ hasdiscrete, needshift
        double precision xshift, xstep, xsh(nx), obj
        common /discrete/ xshift(nx), xstep(nx)

        if(hasdiscrete.AND.needshift) then
            do 10 i=1,nx
                xsh(i)=x(i)+xshift(i)
10          continue
            sh_obj=obj(nx,xsh)
        else
            sh_obj=obj(nx,x)
        endif

        return
        end

c
c   shifted cst(m,nx,x)
c
        double precision function sh_cst(m,n,x)
        double precision x(*)

        parameter(nx=2)

        logical hasdiscrete, needshift
        common /hasdiscrete/ hasdiscrete, needshift
        double precision xshift, xstep, xsh(nx), cst
        common /discrete/ xshift(nx), xstep(nx)

        if(hasdiscrete.AND.needshift) then
            do 10 i=1,nx
                xsh(i)=x(i)+xshift(i)
10          continue
            sh_cst=cst(m,nx,xsh)
        else
            sh_cst=cst(m,nx,x)
        endif

        return
        end

c
c   Unified equality constraints
c
        double precision function cst(m,nx,x)
        double precision x(*)

        parameter(nc=1, neqc=1)
        double precision ceq, cineq, Dmax1

        common /mLag/ mLag
        double precision rleq
        common /rleq/ rleq

c   mLag == 10: relax equality to inequality constraint
        logical hasdiscrete, needshift 
        common /hasdiscrete/ hasdiscrete, needshift

        if(m.LE.neqc) then
            if(mLag.EQ.10) then
                cst=Dmax1(0.0D0,(abs(ceq(m,nx,x))-rleq))
            else
                if(hasdiscrete) then
                    cst=Dmax1(0.0D0,(abs(ceq(m,nx,x))-0.001D0))
                else
                    cst=abs(ceq(m,nx,x))
                endif
            endif
        else if(m.LE.nc) then
            cst=Dmax1(0.0D0,cineq((m-neqc),nx,x))
        else
            print *, "Wrong constraint index in cst"
            stop
        endif

        return
        end
c
c
c   Derivative of unified equality constraints (NOT CORRECT)
c
        subroutine dcst(m,nx,x,y)
        double precision x(*), y(*)

        parameter(nc=1, neqc=1)

        if(m.LE.neqc) then
            call dceq(m,nx,x,y)
        else if(m.LE.nc) then
            call dcineq((m-neqc),nx,x,y)
        else
            print *, "Wrong constraint index in dcst"
            stop
        endif

        return
        end

c
c   shifted dlag(x)
c
        double precision function sh_dlag(x)
        double precision x(*)

        parameter(nx=2, nc=1, neq=nx+nc)

        logical hasdiscrete, needshift
        common /hasdiscrete/ hasdiscrete, needshift
        double precision xshift, xstep, xsh(neq), dlag, dlagRU
        common /discrete/ xshift(nx), xstep(nx)

        logical reuse
        common /reuse/ reuse

        if(hasdiscrete.AND.needshift) then
            do 10 i=1,nx
                xsh(i)=x(i)+xshift(i)
10          continue
            do 20 i=nx+1,neq
                xsh(i)=x(i)
20          continue
            if(reuse) then
                sh_dlag=dlagRU(xsh)
            else
                sh_dlag=dlag(xsh)
            endif
        else
            if(reuse) then
                sh_dlag=dlagRU(x)
            else
                sh_dlag=dlag(x)
            endif
        endif

        return
        end

c
c   dlag(x): Lagrangian function
c       x --- primary and dual variables
c
c     mLag=0: Lag=f(x)+u0*|h(x)|+u1*Max(0,g(x))
c     mLag=1: Lag=f(x)+u0*h(x)+p*h^2(x)/2+{Max(0,u1+p*g(x))^2-u1^2}/(2p)
c     mLag=2: Lag=f(x)+u0*h(x)+p*h^2(x)/2+u1*Max(0,g(x))+p*Max(0,g(x))^2/2
c     mLag=3: Lag=f(x)+u0*|h(x)|+p*h^2(x)/2+u1*Max(0,g(x))+p*Max(0,g(x))^2/2
c     mLag=4: Lag=f(x)+c*Max(0,g(x),|h(x)|)
c     mLag=5: Lag=f(x)+|u0*h(x)|+u1*Max(0,g(x))+p*h^2(x)/2+p*Max(0,g(x))^2/2
c    mLag=10: Lag=f(x)+u0*Max(0,|h(x)|-rleq)+u1*Max(0,g(x))+penalty

        double precision function dlag(x)
        double precision x(*)

        parameter(nx=2, nc=1, neqc0=1)
        parameter(nineqc0=nc-neqc0)

        double precision p, rleq
        common /mLag/ mLag
        common /penalty/ p
        common /rleq/ rleq

        double precision cst, obj, ceq, cineq, Dmax1
        double precision gm, gmax

        double precision slobj
        common /slobj/ slobj

        neqc=neqc0
        nineqc=nineqc0

        dlag=obj(nx,x)/slobj

        if(mLag.EQ.0) then
            do 10 m=1,nc
                dlag=dlag+x(nx+m)*cst(m,nx,x)
10          continue
        else if(mLag.EQ.1) then
            do 20 m=1,neqc
                gm=ceq(m,nx,x)
                dlag=dlag+x(m+nx)*gm+p*gm*gm/2.0D0
20          continue
            do 25 m=1,nineqc
                lx=nx+neqc+m
                gm=p*cineq(m,nx,x)+x(lx)
                if(gm.GT.0.0D0) then
                    dlag=dlag+(gm*gm-x(lx)**2)/(2.0D0*p)
                else
                    dlag=dlag-x(lx)**2/(2.0D0*p)
                endif
25          continue
        else if(mLag.EQ.2) then
            do 30 m=1,neqc
                gm=ceq(m,nx,x)
                dlag=dlag+x(m+nx)*gm+p*gm*gm/2.0D0
30          continue
            do 35 m=1,nineqc
                lx=nx+neqc+m
                gm=Dmax1(0.0D0,cineq(m,nx,x))
                dlag=dlag+gm*x(lx)+p*gm*gm/2.0D0
35          continue
        else if(mLag.EQ.3) then
            do 40 m=1,neqc
                gm=ceq(m,nx,x)
                dlag=dlag+x(m+nx)*abs(gm)+p*gm*gm/2.0D0
40          continue
            do 45 m=1,nineqc
                lx=nx+neqc+m
                gm=Dmax1(0.0D0,cineq(m,nx,x))
                dlag=dlag+gm*x(lx)+p*gm*gm/2.0D0
45          continue
        else if(mLag.EQ.4) then
            gmax=0.0D0
            do 50 m=1,neqc
                gm=abs(ceq(m,nx,x))
                if(gm.GT.gmax) gmax=gm
50          continue
            do 55 m=1,nineqc
                gm=cineq(m,nx,x)
                if(gm.GT.gmax) gmax=gm
55          continue
            dlag=dlag+p*gmax
        else if(mLag.EQ.5) then
            do 60 m=1,neqc
                gm=ceq(m,nx,x)
                dlag=dlag+abs(x(m+nx)*gm)+p*gm*gm/2.0D0
60          continue
            do 65 m=1,nineqc
                lx=nx+neqc+m
                gm=Dmax1(0.0D0,cineq(m,nx,x))
                dlag=dlag+gm*x(lx)+p*gm*gm/2.0D0
65          continue
        else if(mLag.EQ.10) then
            do 70 m=1,neqc
                gm=ceq(m,nx,x)
                dlag=dlag+x(m+nx)*Dmax1(0.0D0,abs(gm)-rleq)+p*gm*gm/2.0D0
70          continue
            do 75 m=1,nineqc
                lx=nx+neqc+m
                gm=Dmax1(0.0D0,cineq(m,nx,x))
                dlag=dlag+gm*x(lx)+p*gm*gm/2.0D0
75          continue
        else
            print *, "Wrong Lagrangian function in dlag()"
            stop
        endif

        return
        end

c
c   shifted lambda_rule(x)
c
        subroutine sh_lambda_rule(x)
        double precision x(*)

        parameter(nx=2, nc=1, neq=nx+nc)

        logical hasdiscrete, needshift
        common /hasdiscrete/ hasdiscrete, needshift
        double precision xshift, xstep, xsh(neq)
        common /discrete/ xshift(nx), xstep(nx)

        if(hasdiscrete.AND.needshift) then
            do 10 i=1,nx
                xsh(i)=x(i)+xshift(i)
10          continue
            do 20 i=nx+1,neq
                xsh(i)=x(i)
20          continue
            call lambda_rule(xsh)
            do 30 i=nx+1,neq
                x(i)=xsh(i)
30          continue
        else
            call lambda_rule(x)
        endif

        return
        end

C
C   lambda_rule: given x, update lambda based on Lagrangian
C                function defined by dlag()

        subroutine lambda_rule(x)
        double precision x(*)

        parameter(nx=2, nc=1, neqc0=1)
        parameter(nineqc0=nc-neqc0, neq=nx+nc)
      
        common /mLag/mLag

        double precision cst, Dmax1, ceq, cineq
        double precision p
        common /penalty/ p

        neqc=neqc0
        nineqc=nineqc0

        if(mLag.EQ.0) then
             do 20 i=nx+1,neq
                x(i)=x(i)+cst((i-nx),nx,x)
 20          continue
        else if(mLag.EQ.1) then
             do 30 m=1,neqc
                x(nx+m)=x(nx+m)+p*ceq(m,nx,x)
 30          continue
             do 35 m=1,nineqc
                lx=nx+neqc+m
                x(lx)=Dmax1(0.0D0,(x(lx)+p*cineq(m,nx,x)))
 35          continue
        else if(mLag.EQ.2) then
             do 40 m=1,neqc
                x(nx+m)=x(nx+m)+p*ceq(m,nx,x)
 40          continue
             do 45 m=1,nineqc
                lx=nx+neqc+m
                x(lx)=x(lx)+p*Dmax1(0.0D0,cineq(m,nx,x))
 45          continue
        else if(mLag.EQ.3) then
             do 50 m=1,neqc
                x(nx+m)=x(nx+m)+p*abs(ceq(m,nx,x))
 50          continue
             do 55 m=1,nineqc
                lx=nx+neqc+m
                x(lx)=x(lx)+p*Dmax1(0.0D0,cineq(m,nx,x))
 55          continue
        endif

        return
        end

c
c   shifted viomax(x)
c
        double precision function sh_viomax(x)
        double precision x(*)

        parameter(nx=2)

        logical hasdiscrete, needshift
        common /hasdiscrete/ hasdiscrete, needshift
        double precision xshift, xstep, xsh(nx), viomax
        common /discrete/ xshift(nx), xstep(nx)

        if(hasdiscrete.AND.needshift) then
            do 10 i=1,nx
                xsh(i)=x(i)+xshift(i)
10          continue
            sh_viomax=viomax(xsh)
        else
            sh_viomax=viomax(x)
        endif

        return
        end

c
c   viomax(x): maximal violation of constraints
c
        double precision function viomax(x)
        double precision x(*)

        parameter(nc=1)

        double precision cst, val

        viomax=0.0D0

        do 10 m=1,nc
            val=cst(m,nx,x)
            if(val.GT.viomax) viomax=val
10      continue

        return
        end

c
c   shifted ocst(m,nx,x)
c
        double precision function sh_ocst(m,n,x)
        double precision x(*)

        parameter(nx=2)

        logical hasdiscrete, needshift
        common /hasdiscrete/ hasdiscrete, needshift
        double precision xshift, xstep, xsh(nx), ocst
        common /discrete/ xshift(nx), xstep(nx)

        if(hasdiscrete.AND.needshift) then
            do 10 i=1,nx
                xsh(i)=x(i)+xshift(i)
10          continue
            sh_ocst=ocst(m,nx,xsh)
        else
            sh_ocst=ocst(m,nx,x)
        endif

        return
        end

c
c   Unified original equality constraints
c
        double precision function ocst(m,nx,x)
        double precision x(*)

        parameter(nc=1, neqc=1)
        double precision ceq, cineq, Dmax1

        logical hasdiscrete, needshift
        common /hasdiscrete/ hasdiscrete, needshift

        if(m.LE.neqc) then
            if(hasdiscrete) then
                ocst=Dmax1(0.0D0,(abs(ceq(m,nx,x))-0.001D0))
            else
                ocst=abs(ceq(m,nx,x))
            endif
        else if(m.LE.nc) then
            ocst=Dmax1(0.0D0,cineq((m-neqc),nx,x))
        else
            print *, "Wrong constraint index in ocst"
            stop
        endif

        return
        end

c
c   shifted oviomax(x)
c
        double precision function sh_oviomax(x)
        double precision x(*)

        parameter(nx=2)

        logical hasdiscrete, needshift
        common /hasdiscrete/ hasdiscrete, needshift
        double precision xshift, xstep, xsh(nx), oviomax
        common /discrete/ xshift(nx), xstep(nx)

        if(hasdiscrete.AND.needshift) then
            do 10 i=1,nx
                xsh(i)=x(i)+xshift(i)
10          continue
            sh_oviomax=oviomax(xsh)
        else
            sh_oviomax=oviomax(x)
        endif

        return
        end

c
c   oviomax(x): maximal violation of original equality constraints
c
        double precision function oviomax(x)
        double precision x(*)

        parameter(nc=1)

        double precision ocst, val

        oviomax=0.0D0

        do 10 m=1,nc
            val=ocst(m,nx,x)
            if(val.GT.oviomax) oviomax=val
10      continue

        return
        end

c
c   subroutine dxinfo(n,dx,xshift)
c       n: dimension of xshift
c       dx: discrete step size
c       xshift: get actual variable by x + xshift
c
        subroutine dxinfo(n,dx,xshift)
        double precision dx(*), xshift(*)

        parameter(nx=2, ns=10000)

        integer isdx 
        common /isdx/ isdx(nx)
        double precision ndsample
        common /ndsample/ ndsample(nx)
        logical hasdiscrete, needshift
        common /hasdiscrete/ hasdiscrete, needshift
        double precision xx, dis, dismin, xirange
        double precision bl(nx), bu(nx), xbest(nx)
        double precision obj, oviomax
        double precision nis, Dint

        call xbound(nx,bl,bu)
        call xbestsolu(nx,xbest)

        print *, "number of discretized points per unit =", ns
c        print *, "Best obj =", obj(nx,xbest), "   vioMax =", oviomax(xbest)

        do 20 i=1,nx
            xirange=bu(i)-bl(i)
            if(xirange.LT.1.0D0) then
                nis=ns
            else
                nis=Dint(xirange*ns+1.0D-6)
            endif
            ndsample(i)=nis
            if(isdx(i).EQ.1) then
                dx(i)=xirange/(nis-1.0D0)
            else
                dx(i)=0.0D0
            endif
20      continue

        needshift=.FALSE.
        do 30 i=1,nx
            xshift(i)=0.0D0
30      continue
        return

        do 5 i=1,nx
            if(isdx(i).EQ.1) then
                jx=0
                dismin=1.0D20
                do 10 j=0,ndsample(i)-1
                    xx=bl(i)+j*dx(i)
                    dis=abs(xx-xbest(i))
                    if(dis.LT.dismin) then
                        jx=j
                        dismin=dis
                    endif
10              continue
                xshift(i)=xbest(i)-(bl(i)+jx*dx(i))
                if(abs(xshift(i)).GT.1.0D-6) needshift=.TRUE.
            else
                xshift(i)=0.0D0
            endif
5       continue

c        print *, "hasdiscrete = ", hasdiscrete, "   needshift = ", needshift
c        print *, " "

        call flush(6)

        return
        end

c
c   gdesc_inx(niter,x): simple gradient descent 
c       niter: the number of iterations
c       x: input and output x
c       

        subroutine gdesc_inx(niter,xin)
        double precision xin(*)

        parameter(nx=2, nc=1)
        parameter(neq=nx+nc) 

        double precision dlag, viomax 
        double precision x(neq), xdr(nx), drmax, dterm
        double precision xgo, xstep, xvio, xinvio
        double precision dx, xt(neq), xlag, xtlag
        data dx/1.0D-6/

        do 2 i=1,neq
            x(i)=xin(i)
2       continue

        do 5 i=nx+1,neq
            xt(i)=x(i)
5       continue

        do 50 iter=1,niter
            do 10 i=1,nx
               xt(i)=x(i)
10          continue

            drmax=0.0D0
            xlag=dlag(x)
            do 20 i=1,nx
               xt(i)=x(i)+dx
               xtlag=dlag(xt)
               dterm=(xlag-xtlag)/dx
               xdr(i)=dterm
               dterm=abs(dterm)
               if(dterm.GT.drmax) drmax=dterm
               xt(i)=x(i)
20          continue
            if(drmax.LT.1.0D-5) goto 100
            if(drmax.GT.1.0D0) then
               do 25 i=1,nx
                  xdr(i)=xdr(i)/drmax
25             continue
            endif

            drmax=(-1.0D20)
            do 30 m=0,5
               xstep=0.01D0*(2**m)
               do 35 i=1,nx
                  xt(i)=x(i)+xstep*xdr(i)
35             continue
               xtlag=dlag(xt)
               dterm=xlag-xtlag
               if(dterm.GT.drmax) then
                  drmax=dterm
                  xgo=xstep
               endif
30          continue
            if(drmax.GT.0.0D0) then
               do 40 i=1,nx
                   x(i)=x(i)+xgo*xdr(i)
40             continue
            endif
50      continue

100     xvio=viomax(x) 
        xinvio=viomax(xin)
        if(xvio.LT.1.0D-4.OR.xvio.LT.(2.0D0*xinvio)) then
           do 60 i=1,neq
              xin(i)=x(i)
60         continue
        endif

        return
        end

c
c   dlagRU(x): Lagrangian function
c       x --- primary and dual variables
c
        double precision function dlagRU(x)
        double precision x(*)

        parameter(nx=2, nc=1, neqc0=1)
        parameter(neq=nx+nc, nineqc0=nc-neqc0)

        double precision p, rleq
        common /mLag/ mLag
        common /penalty/ p
        common /rleq/ rleq

        double precision cst, obj, ceq, cineq, Dmax1
        double precision gm, gmax

C    reuse index
        parameter(nprt=3, nchd=3)
        parameter(ntot=nprt+nchd)

        integer icx,ix
        common /RUindex/ icx, ix(ntot)
        double precision objRU, cstRU 
        common /RUdata/ objRU(ntot), cstRU(ntot,nc+1)

        neqc=neqc0
        nineqc=nineqc0

        if(ix(icx).EQ.0) objRU(icx)=obj(nx,x)
        dlagRU=objRU(icx)

        if(mLag.EQ.0) then
            do 10 m=1,nc
                dlag=dlag+x(nx+m)*cst(m,nx,x)
10          continue
        else if(mLag.EQ.1) then
            do 20 m=1,neqc
                gm=ceq(m,nx,x)
                dlag=dlag+x(m+nx)*gm+p*gm*gm/2.0D0
20          continue
            do 25 m=1,nineqc
                lx=nx+neqc+m
                gm=p*cineq(m,nx,x)+x(lx)
                if(gm.GT.0.0D0) then
                    dlag=dlag+(gm*gm-x(lx)**2)/(2.0D0*p)
                else
                    dlag=dlag-x(lx)**2/(2.0D0*p)
                endif
25          continue
        else if(mLag.EQ.2) then
            do 30 m=1,neqc
                gm=ceq(m,nx,x)
                dlag=dlag+x(m+nx)*gm+p*gm*gm/2.0D0
30          continue
            do 35 m=1,nineqc
                lx=nx+neqc+m
                gm=Dmax1(0.0D0,cineq(m,nx,x))
                dlag=dlag+gm*x(lx)+p*gm*gm/2.0D0
35          continue
        else if(mLag.EQ.3) then
            do 40 m=1,neqc
                if(ix(icx).EQ.0) cstRU(icx,m)=ceq(m,nx,x)
                gm=cstRU(icx,m)
                dlagRU=dlagRU+x(m+nx)*abs(gm)+p*gm*gm/2.0D0
40          continue
            do 45 m=1,nineqc
                if(ix(icx).EQ.0) cstRU(icx,neqc+m)=Dmax1(0.0D0,cineq(m,nx,x))
                lx=nx+neqc+m
                gm=cstRU(icx,neqc+m)
                dlagRU=dlagRU+gm*x(lx)+p*gm*gm/2.0D0
45          continue
            ix(icx)=1
        else if(mLag.EQ.4) then
            gmax=0.0D0
            do 50 m=1,neqc
                gm=abs(ceq(m,nx,x))
                if(gm.GT.gmax) gmax=gm
50          continue
            do 55 m=1,nineqc
                gm=cineq(m,nx,x)
                if(gm.GT.gmax) gmax=gm
55          continue
            dlag=dlag+p*gmax
        else if(mLag.EQ.5) then
            do 60 m=1,neqc
                gm=ceq(m,nx,x)
                dlag=dlag+abs(x(m+nx)*gm)+p*gm*gm/2.0D0
60          continue
            do 65 m=1,nineqc
                lx=nx+neqc+m
                gm=Dmax1(0.0D0,cineq(m,nx,x))
                dlag=dlag+gm*x(lx)+p*gm*gm/2.0D0
65          continue
        else if(mLag.EQ.10) then
            do 70 m=1,neqc
                gm=ceq(m,nx,x)
                dlag=dlag+x(m+nx)*Dmax1(0.0D0,abs(gm)-rleq)+p*gm*gm/2.0D0
70          continue
            do 75 m=1,nineqc
                lx=nx+neqc+m
                gm=Dmax1(0.0D0,cineq(m,nx,x))
                dlag=dlag+gm*x(lx)+p*gm*gm/2.0D0
75          continue
        else
            print *, "Wrong Lagrangian function in dlag()"
            stop
        endif

        return
        end

