c
c   xpi(iscauchy,i,xi,vmi,lbi,ubi): neighborhood for x variable
c       iscauchy: iscauchy=0, local neighbor; iscauchy=1, cauchy distribution
c       i: index for xi
c       xi: x(i),  vmi=vm(i),  lbi=lb(i),  and ubi=ub(i) 
c
        double precision function xpi(iscauchy,i,xi,vmi,lbi,ubi)
        integer iscauchy, i
        double precision xi, vmi, lbi, ubi

        parameter(nx=2)

        integer isdx 
        common /isdx/ isdx(nx)
        double precision ndsample
        common /ndsample/ ndsample(nx)
        double precision xshift, xstep
        common /discrete/ xshift(nx), xstep(nx)

        real RANMAR, rd
        double precision u1,u2,gasdev
        double precision nxstep, Dint

C    generate points close to current point by uniform distribution
        if(iscauchy.EQ.0) then
            rd=RANMAR()*2.0-1.0
            if(isdx(i).EQ.0) then
                xpi=xi+rd*vmi
                if(xpi.LT.lbi.OR.xpi.GT.ubi) xpi=lbi+(ubi-lbi)*RANMAR()
            else
                nxstep=Dint(rd*vmi/xstep(i))
                if(abs(nxstep).LT.1.0D-8) then
                    if(rd.GT.0.0) then
                        nxstep=1.0D0
                    else
                        nxstep=(-1.0D0)
                    endif
                endif
                xpi=xi+nxstep*xstep(i)
                if(xpi.LT.(lbi-1.0D-6).OR.xpi.GT.(ubi+1.0D-6)) then
                    xpi=lbi+Dint(RANMAR()*ndsample(i))*xstep(i)
                endif
            endif
            return
        endif

C    generate points close to current point with Gaussian
        if(iscauchy.EQ.2) then
            rd=gasdev()
            if(isdx(i).EQ.0) then
                xpi=xi+rd*vmi
                if(xpi.LT.lbi.OR.xpi.GT.ubi) xpi=lbi+(ubi-lbi)*RANMAR()
            else
                nxstep=Dint(rd*vmi/xstep(i))
                if(abs(nxstep).LT.1.0D-8) then
                    if(rd.GT.0.0) then
                        nxstep=1.0D0
                    else
                        nxstep=(-1.0D0)
                    endif
                endif
                xpi=xi+nxstep*xstep(i)
                if(xpi.LT.(lbi-1.0D-6).OR.xpi.GT.(ubi+1.0D-6)) then
                    xpi=lbi+Dint(RANMAR()*ndsample(i))*xstep(i)
                endif
            endif
            return
        endif

c    generate points with Log SA
        if(iscauchy.EQ.3) then
            rd=RANMAR()*2.0-1.0
c            xpi=(xi-lbi)*(10.0D0**(rd*vmi))+lbi
            xpi=xi*(10.0D0**(rd*vmi))
            if(isdx(i).EQ.0) then
                if(xpi.LT.lbi.OR.xpi.GT.ubi) xpi=lbi+(ubi-lbi)*RANMAR()
            else
                xpi=Dint((xpi-lbi)/xstep(i))*xstep(i)+lbi
                if(xpi.LT.(lbi-1.0D-6).OR.xpi.GT.(ubi+1.0D-6)) then
                    xpi=lbi+Dint(RANMAR()*ndsample(i))*xstep(i)
                endif
            endif
            return
        endif

c    generate points with Cauchy distribution
        if(abs(ubi-lbi)/(abs(ubi)+abs(lbi)).LT.1.0D-8) then
            xpi=0.5D0*(ubi+lbi)
            return
        endif
100     u1=RANMAR()*2.0D0-1.0D0
        u2=RANMAR()*2.0D0-1.0D0
        if((u1*u1+u2*u2).GE.1.0D0) goto 100
        if(isdx(i).EQ.0) then
            xpi=xi+vmi*(u2/u1)
            if(xpi.LT.lbi.OR.xpi.GT.ubi) goto 100
        else
            nxstep=Dint(vmi*(u2/u1)/xstep(i))
            if(abs(nxstep).LT.1.0D-8) then
                if(u1.GT.0.0D0) then
                    nxstep=1.0D0
                else
                    nxstep=(-1.0D0)
                endif
            endif
            xpi=xi+nxstep*xstep(i)
            if(xpi.LT.(lbi-1.0D-6).OR.xpi.GT.(ubi+1.0D-6)) goto 100
        endif

        return
        end

c
c   gasdev(): return a Gaussian random variable
c
        double precision function gasdev()

        integer iGset
        double precision rGset
        common /GaussSet/ iGset, rGset

        real RANMAR
        double precision v1, v2, fac, rsq

        if(iGset.EQ.0) then
10          v1=RANMAR()*2.0D0-1.0D0
            v2=RANMAR()*2.0D0-1.0D0
            rsq=v1*v1+v2*v2
            if(rsq.GE.1.0D0.OR.rsq.EQ.0.0D0) goto 10
            fac=sqrt(-2.0D0*log(rsq)/rsq)
            rGset=v1*fac
            iGset=1
            gasdev=v2*fac
        else
            iGset=0
            gasdev=rGset
        endif

        return
        end

c
c   by_DLM(): discrete Lagrangian search
c
        subroutine by_DLM(msol,x,xobj,xvio,ni_used)
        integer msol, ni_used
        double precision x(*), xobj, xvio

        parameter(nx=2, nc=1)
        parameter(neq=nx+nc, NS=50, maxIt=1000)

        integer iscauchy, nlchange, ni, nacp(nx)

        double precision vtol, f, fp, fmin, fval, ratio, vl
        double precision xpi, sh_oviomax, sh_obj, sh_ocst
        double precision lb(nx), ub(nx), vm(nx)
        double precision xopt(neq), xp(neq)

        logical xchange

        vtol=1.0D-5

        do 1 i=nx+1,neq
            x(i)=0.0D0
1       continue
C        msol=0
C        return

        if(sh_oviomax(x).LT.vtol) then
            fmin=sh_obj(nx,x)
            do 5 i=1,neq
                xopt(i)=x(i)
5           continue
        else
            fmin=1.0D+20
            do 6 i=1,neq
                xopt(i)=0.0D0
6           continue
        endif

        call fcn(nx,x,f)
        call xbound(nx,lb,ub)

        do 15 i=1,nx
            nacp(i)=0
            vm(i)=0.01D0*(ub(i)-lb(i))
15      continue

        ni=0
        nlchange=0

100     nnew=0

        do 10 i=1,neq
            xp(i)=x(i)
10      continue

        xchange=.FALSE.
        do 200 iter=1,NS
          do 20 i=1,nx
            do 25 iscauchy=0,1
                xp(i)=xpi(iscauchy,i,x(i),vm(i),lb(i),ub(i))
                call fcn(nx,xp,fp)
                if(fp.LT.f) then
                    f=fp
                    x(i)=xp(i)
                    if(iscauchy.EQ.0) nacp(i)=nacp(i)+1
                    xchange=.TRUE.
                    if(sh_oviomax(x).LT.vtol) then
                        fval=sh_obj(nx,x)
                        if(fval.LT.fmin) then
                            do 30 j=1,neq
                                xopt(j)=x(j)
30                          continue
                            fmin=fval
                            nnew=nnew+1
                        endif
                    endif
                else
                    xp(i)=x(i)
                endif
25          continue
20        continue
200     continue

        ni=ni+1
        do 40 i=1,nx
            ratio=dfloat(nacp(i))/dfloat(NS)
            if(ratio.GT.0.2D0) then
                vm(i)=1.5D0*vm(i)
            else if(ratio.LT.0.1D0) then
                vm(i)=vm(i)/1.5D0
            endif
            if(vm(i).GT.(ub(i)-lb(i))) vm(i)=ub(i)-lb(i)
40      continue
        do 45 i=1,nx
            nacp(i)=0
45      continue

        if(ni.GE.maxIt.OR.nlchange.GE.10) goto 500
        if(xchange) then
            nlchange=0
        else
            nlchange=nlchange+1
            do 50 i=1,nc
                vl=sh_ocst(i,nx,x)
                x(i+nx)=x(i+nx)+vl
50          continue
        endif
        if(nnew.GE.1) then
            do 70 i=1,nx
                x(i)=xopt(i)
70          continue 
        endif
        goto 100

500     ni_used=ni
        xvio=sh_oviomax(xopt)
        if(xvio.LT.vtol) then
            msol=1
            xobj=sh_obj(nx,xopt)
            do 60 i=1,neq
                x(i)=xopt(i)
60          continue
        else
            msol=0
            xobj=sh_obj(nx,x)
            xvio=sh_oviomax(x)
        endif

        return
        end


