C  population-based SA with stochastic Lambda rule
C
C    After probablity descent in x for a number of steps, do
C    probability ascent in lambda

      SUBROUTINE pSA(rlambda,
     1              N,xx,MAX,RT,EPS,NS,NT,NEPS,MAXEVL,LB,UB,C,IPRINT,
     2              ISEED1,ISEED2,T,VM,XOPT,FOPT,NACC,NFCNEV,NOBDS,IER,
     3              FSTAR,xxp,NACP)

C  Parameter definition
      parameter(nx=4, nc=5)
      parameter(neq=nx+nc)

      parameter(nprt=10, nchd=10)
      parameter(ntot=nprt+nchd)

C  for population
      double precision xx(nprt,neq), xxp(ntot,neq)
      double precision ff(nprt), x(neq), xp(neq)
      double precision avgvio(nc)

C  Type all external variables.
      DOUBLE PRECISION  LB(*), UB(*), C(*), VM(*), FSTAR(*),
     1                  XOPT(*), T, EPS, RT, FOPT
      INTEGER  NACP(*), N, NS, NT, NEPS, NACC, MAXEVL, IPRINT,
     1         NOBDS, IER, NFCNEV, ISEED1, ISEED2
      LOGICAL  MAX

C  Type all internal variables.
      DOUBLE PRECISION  F, FP, P, PP, RATIO
      INTEGER  NNEW, H, I, J, M
      LOGICAL  QUIT

C  Type all functions.
      DOUBLE PRECISION  EXPREP
      REAL  RANMAR

C  For cauchy distribution
      integer iscauchy
      double precision rateup, ratelow

C  For lambda
      integer ndecl(nc), nincl(nc)

C  For lambda's determinimistic rule
      integer neq
      double precision  fmin, fval, vtol, vl, rlambda(*)
      common /vtol/ vtol
      double precision  sh_cst, sh_viomax, sh_obj, xpi
      double precision  t0, xvio

      integer nreSA

      logical reuse
      common /reuse/ reuse
      integer icx,ix 
      common /RUindex/ icx, ix(ntot)

      reuse=.TRUE.
      do 150 i=1,ntot
         ix(i)=0
150   continue

      iscauchy=1
      nreSA=0
      t0=t

      if(iscauchy.EQ.0) then
         rateup=0.6D0
         ratelow=0.4D0
      else
         rateup=0.3D0
         ratelow=0.2D0
      endif

C  Initialize the random number generator RANMAR.
      CALL RMARIN(ISEED1,ISEED2)

C  Set initial values.
      NACC = 0
      NOBDS = 0
      NFCNEV = 0
      IER = 99

      DO 5, I = 1, N
         NACP(I) = 0
5     CONTINUE

      DO 10, I = 1, NEPS
         FSTAR(I) = 1.0D+20
10    CONTINUE 

C  If the initial temperature is not positive, notify the user and 
C  return to the calling routine.  
      IF (T .LE. 0.0) THEN
         WRITE(*,'(/,''  THE INITIAL TEMPERATURE IS NOT POSITIVE. ''
     1             /,''  RESET THE VARIABLE T. ''/)')
         IER = 3
         RETURN
      END IF

C  If the initial value is out of bounds, notify the user and return
C  to the calling routine.
      do 15 k=1, nprt
         do 20 j=1, nx
            IF ((xx(k,j) .GT. UB(j)) .OR. (xx(k,j) .LT. LB(j))) THEN
               CALL PRT1
               IER = 2
               RETURN
            END IF
20       continue
15    continue

C  Evaluate the function with input X and return value as F.
      FOPT = 0.0D0
      FSTAR(1) = 0.0D0 
      fmin=1.0D20
      do 25 j=1,neq
         xopt(j)=0.0D0
25    continue

      do 40 k=1,nprt
          do 45 j=1,neq
              x(j)=xx(k,j)
45        continue
          if(reuse) icx=k
          call fcn(nx,x,ff(k))
          if(sh_viomax(x).LT.vtol) then
             fval=sh_obj(n,x)
             if(fval.LT.fmin) then
                fmin=fval
                do 42 j=1,neq
                   xopt(j)=x(j)
42              continue
                fopt=ff(k)
             endif
          endif
40    continue

C  Determine lambda's VIM by degree of violation
      do 30 j=1,nc
         rlambda(j)=0.01D0
c         vm(j+n)=rlambda(j)*sh_cst(j,n,x)
30    continue

C  Start the main loop. Note that it terminates if (i) the algorithm
C  succesfully optimizes the function or (ii) there are too many
C  function evaluations (more than MAXEVL).
100   NNEW = 0

      do 35 i=1,nc
         ndecl(i)=0
         nincl(i)=0
35    continue

      DO 400, M = 1, NT

         do 70 k=1, nprt
            do 75 i=nx+1, neq
               xxp(k,i)=xx(k,i)
               xxp(k+nprt,i)=xx(k,i)
75          continue
70       continue

         DO 300, J = 1, NS
            DO 200, H = 1, N

C  Generate trial population xxp from xx. Note use of VM to choose XP.
               do 210 k=1, nprt
                  do 215 i=1, nx
                     xxp(k,i)=xx(k,i)
                     xxp(k+nprt,i)=xx(k,i)
215               continue
210            continue

               do 250 k=1, nprt
                  if(iscauchy.EQ.0) then
                      xxp(k+nprt,h)=xpi(0,h,xx(k,h),vm(h),lb(h),ub(h))
                  else
                      xxp(k+nprt,h)=xpi(1,h,xx(k,h),vm(h),lb(h),ub(h))
                  endif
250            continue

               call selection(ff,xx,xxp,t,nhacp,fmin,xopt,fopt,nnew)

               NACC = NACC + nhacp
               NACP(H) = NACP(H) + nhacp

200         CONTINUE
300      CONTINUE

C  Adjust VM so that approximately half of all evaluations are accepted.
         DO 310, I = 1, N
            RATIO = DFLOAT(NACP(I)) /DFLOAT(NS*nprt)
            IF (RATIO .GT. rateup) THEN
               vm(I) = vm(I)*(1. + 3.5D0*C(I)*(RATIO - rateup)/(1.0D0-rateup))
            ELSE IF (RATIO .LT. ratelow) THEN
               vm(I) = vm(I)/(1. + C(I)*((ratelow - RATIO)/ratelow))
            END IF
            IF (vm(I) .GT. (UB(I)-LB(I))) vm(I) = UB(I) - LB(I)
310      CONTINUE

         do 80 i=1,nc
            avgvio(i)=0.0D0
            do 85 k=1,nprt
               do 90 j=1,nx
                  x(j)=xx(k,j)
90             continue
               vl=sh_cst(i,n,x)
               avgvio(i)=avgvio(i)+vl
85          continue
            avgvio(i)=avgvio(i)/nc
80       continue

         do 50 i=1,nc
            vl=avgvio(i)
            if(vl.LT.0.01D0*t) then
                ndecl(i)=ndecl(i)+1
                nincl(i)=0
                if(ndecl(i).GE.3) then
                    ndecl(i)=0
                    rlambda(i)=rlambda(i)*0.8D0
                    if(rlambda(i).LT.0.001D0) rlambda(i)=0.001D0
                endif
            else if(vl.GT.1.0D0*t) then
                nincl(i)=nincl(i)+1
                ndecl(i)=0
                if(nincl(i).GE.3) then
                    nincl(i)=0
                    rlambda(i)=rlambda(i)*1.25D0
                endif
            else
                ndecl(i)=0
                nincl(i)=0
            endif
c            vm(i+n)=rlambda(i)*vl
50       continue

         do 110 k=1, nprt
            f=ff(k)
            do 65 h=1,neq
               x(h)=xx(k,h)
               xp(h)=x(h)
65          continue

            do 60 h=n+1,neq
                vl=sh_cst((h-n),n,x)
                if(vl.LT.vtol) goto 60
                xp(h)=x(h)+(ranmar()*2.-1.)*rlambda(h-n)*vl
                if(xp(h).LT.0.0D0) xp(h)=0.0D0

                icx=k
                call fcn(n,xp,fp)
                if(fp.GT.f) then
                   x(h)=xp(h)
                   f=fp
                else
                   p=exprep((fp-f)/T)
                   pp=ranmar()
                   if(pp.LT.p) then
                      x(h)=xp(h)
                      f=fp
                   else
                      xp(h)=x(h)
                   endif
                endif
60          continue

            ff(k)=f
            do 115 h=n+1,neq
                xx(k,h)=x(h)
115         continue
110      continue

         DO 120, I = 1, N
            NACP(I) = 0
120      CONTINUE
400   CONTINUE

      do 55 i=1,n
         vm(i)=1.0D0
55    continue

c      print *, " "
c      print *, "Current T = ", T
c      print *, "obj = ", sh_obj(n,x),"    maximal violation=", sh_viomax(x)
c      print *, "x = ", (x(i),i=1,neq)

C  Check termination criteria.
      QUIT = .FALSE.
      FSTAR(1) = F
      IF ((FOPT - FSTAR(1)) .LE. EPS) QUIT = .TRUE.
      DO 410, I = 1, NEPS
         IF (ABS(F - FSTAR(I)) .GT. EPS) QUIT = .FALSE.
410   CONTINUE

      if(t.LT.1.0D-6.OR.quit) then
         xvio=sh_viomax(x)
         if(xvio.GT.vtol.AND.nreSA.GT.0) then
            nreSA=nreSA-1
            quit=.FALSE.
            t=xvio*100.0D0
            if(t.GT.t0) t=t0
            print *, "Re-annealing from T =", t
            goto 100
         else
            quit=.TRUE.
         endif
      endif

C  Terminate SA if appropriate.
      IF (QUIT) THEN
         IER = 0
         RETURN
      END IF

C  If termination criteria is not met, prepare for another loop.
      T = RT*T
      DO 430, I = NEPS, 2, -1
         FSTAR(I) = FSTAR(I-1)
430   CONTINUE

c      if(nnew.GE.1) then
c         F = FOPT
c         DO 440, I = 1, N
c            X(I) = XOPT(I)
c440      CONTINUE
c      endif

C  Loop again.
      GO TO 100

      END

C
C selection procedure
C
      subroutine selection(ff,xx,xxp,t,nhacp,fmin,xopt,fopt,nnew)

      parameter(nx=4, nc=5)
      parameter(neq=nx+nc)

      parameter(nprt=10, nchd=10)
      parameter(ntot=nprt+nchd)

C  for population
      double precision fmin, fopt, xopt(*)
      double precision ff(*), xx(nprt,neq), xxp(ntot,neq), t

      REAL RANMAR
      DOUBLE PRECISION EXPREP, sh_viomax, sh_obj
      double precision x(neq), xl(neq), f(ntot), p(ntot)
      double precision fval, psum, pp

      double precision vtol
      common /vtol/ vtol
      logical newP

      logical reuse
      common /reuse/ reuse
      integer icx,ix 
      common /RUindex/ icx, ix(ntot)
      double precision objRU, cstRU, objtmp(nprt), csttmp(nprt,nc)
      common /RUdata/ objRU(ntot), cstRU(ntot,nc)

      nhacp=0

      do 10 l=1, nprt
         do 35 j=nx+1, neq
            xl(j)=xxp(l,j)
35       continue
         psum=0.0D0
         do 20 k=1, ntot
            if(l.EQ.k) then
               f(k)=ff(k)
            else
               do 30 j=1, nx
                  xl(j)=xxp(k,j)
30             continue
               if(reuse) icx=k
               call fcn(nx,xl,f(k))
            endif
            p(k)=EXPREP(-(f(k)-f(1))/t)
            psum=psum+p(k)
20       continue
         do 40 k=1, ntot
            p(k)=p(k)/psum
40       continue
         
         pp=RANMAR()
         newP=.FALSE.
         do 50 k=1,ntot
             if(pp.LE.p(k)) then
                 kk=k
                 do 60 j=1, nx
                    xx(l,j)=xxp(kk,j)
60               continue
                 ff(l)=f(kk)

                 if(reuse) then
                     ix(l)=1
                     objtmp(l)=objRU(kk)
                     do 65 j=1, nc
                        csttmp(l,j)=cstRU(kk,j)
65                   continue
                 endif

                 if(kk.GT.nprt) then
                    nhacp=nhacp+1
                    newP=.TRUE.
                 endif
                 goto 100
             else
                 p(k+1)=p(k)+p(k+1)
             endif
50       continue
         print *, "wrong in selection"
         stop

C  If greater than any other point, record as new optimum.
100      if(newP) then
             do 90 j=1, neq
                 x(j)=xx(l,j)
90           continue
             if(sh_viomax(x).LT.vtol) then
                fval=sh_obj(nx,x)
                if(fval.LT.fmin) then
                    do 95 i=1, neq
                       xopt(i)=x(i) 
95                  continue
                    fmin=fval
                    fopt=f(kk)
                    nnew=nnew+1
                end if
             end if
         endif
10    continue

      if(reuse) then
         do 70 k=nprt+1,ntot
            ix(k)=0
70       continue
         do 80 k=1,nprt
            objRU(k)=objtmp(k)
            do 85 j=1,nc
               cstRU(k,j)=csttmp(k,j)
85          continue
80       continue
      endif

      return
      end

c
c   print 2-D matrix
c
      subroutine mtxprint(m,n,x)
      double precision x(m,n)

      do 10 i=1,m
         print *,(x(i,j),j=1,n)
10    continue
      
      return
      end

