C  SA with stochastic Lambda rule
C
C    After probablity descent in x for a number of steps, do
C    probability ascent in lambda

      SUBROUTINE SA(NEQ,NC,rlambda,
     1              N,X,MAX,RT,EPS,NS,NT,NEPS,MAXEVL,LB,UB,C,IPRINT,
     2              ISEED1,ISEED2,T,VM,XOPT,FOPT,NACC,NFCNEV,NOBDS,IER,
     3              FSTAR,XP,NACP)

C  Type all external variables.
      DOUBLE PRECISION  X(*), LB(*), UB(*), C(*), VM(*), FSTAR(*),
     1                  XOPT(*), XP(*), T, EPS, RT, FOPT
      INTEGER  NACP(*), N, NS, NT, NEPS, NACC, MAXEVL, IPRINT,
     1         NOBDS, IER, NFCNEV, ISEED1, ISEED2
      LOGICAL  MAX

C  Type all internal variables.
      DOUBLE PRECISION  F, FP, P, PP, RATIO
      INTEGER  NUP, NDOWN, NREJ, NNEW, LNOBDS, H, I, J, M
      LOGICAL  QUIT

C  Type all functions.
      DOUBLE PRECISION  EXPREP
      REAL  RANMAR, rtp

C  For cauchy distribution
      integer iscauchy
      double precision rateup, ratelow

C  For lambda
      parameter(nnc=1)
      integer ndecl(nnc+1), nincl(nnc+1)
      integer numnov(nnc+1), numv(nnc+1)

C  For lambda's determinimistic rule
      integer neq
      double precision  fmin, fval, vtol, vl, rlambda(*)
      common /vtol/ vtol
      double precision  sh_cst, sh_viomax, sh_obj, xpi
      double precision  t0, xvio, dtf

C  Count number of iteration at scale NT
      integer nSAIter
      common /nSAIter/ nSAIter

c penalty methods
      common /mPen/ mPen

C  Predefined neighborhood size, varied with temperature
      parameter(nx=2)
      double precision vm0(nx)

C  static or dynamic neighborhood
C      staticNeb=.false. for dynamic neighbors

      logical staticNeb

c  parameters for Hastings' rule and Tsallis rule
      integer qH
      double precision qT, qT2

C  nreSA: number of reannealing time
      integer nreSA

      nreSA=0
      t0=t

C
C   parameter values for Hastings' and Tsallis rules
c      qH >= 1 and qT > 1
c  acceptance probability for cSA
c      iaccept=0: logistic
c      iaccept=1: Metropolis
c      iaccept=2: Hastings
c      iaccept=3: Tsallis

      iaccept=1
      if(iaccept.EQ.2) qH=2

c  distribution to generate random trial points
c      iscauchy=0: uniform; 
c      iscauchy=1: Cauchy; 
c      iscauchy=2: Gaussian
c      iscauchy=3: Log SA neighborhood
c      staticNeb=.TRUE.

      staticNeb=.FALSE.
      iscauchy=1

      if(iscauchy.EQ.0.OR.iscauchy.EQ.2) then
         rateup=0.6D0
         ratelow=0.4D0
      else if(iscauchy.EQ.1) then
         rateup=0.3D0
         ratelow=0.2D0
      else if(iscauchy.EQ.3) then
         rateup=0.2D0
         ratelow=0.05D0
      else
         print *, "wrong distribution"
         stop
      endif

C  Initialize the random number generator RANMAR.
      CALL RMARIN(ISEED1,ISEED2)

C  Set initial values.
      NACC = 0
      NOBDS = 0
      NFCNEV = 0
      IER = 99

      DO 10, I = 1, N
         NACP(I) = 0
10    CONTINUE

      DO 20, I = 1, NEPS
         FSTAR(I) = 1.0D+20
20    CONTINUE 

C  If the initial temperature is not positive, notify the user and 
C  return to the calling routine.  
      IF (T .LE. 0.0) THEN
         WRITE(*,'(/,''  THE INITIAL TEMPERATURE IS NOT POSITIVE. ''
     1             /,''  RESET THE VARIABLE T. ''/)')
         IER = 3
         RETURN
      END IF

C  If the initial value is out of bounds, notify the user and return
C  to the calling routine.
      DO 30, I = 1, N
         IF ((X(I) .GT. UB(I)) .OR. (X(I) .LT. LB(I))) THEN
             print *, "x(",I,") is outside the bounds"
c            CALL PRT1
c            IER = 2
c            RETURN
         END IF
30    CONTINUE

C  Evaluate the function with input X and return value as F.
      CALL FCN(N,X,F)

C  If the function is to be minimized, switch the sign of the function.
C  Note that all intermediate and final output switches the sign back
C  to eliminate any possible confusion for the user.
      IF(.NOT. MAX) F = -F
      NFCNEV = NFCNEV + 1
      FOPT = F
      FSTAR(1) = F
      IF(IPRINT .GE. 1) CALL PRT2(MAX,NEQ,X,F)

      if(sh_viomax(x).LT.vtol) then
         fmin=sh_obj(n,x)
         do 8 i=1,neq
            xopt(i)=x(i)
8        continue
      else
         fmin=1.0D+20
         do 9 i=1,neq
            xopt(i)=0.0D0
9        continue
      endif

C  determine VM for LogSA and predefined neighbors
      if(iscauchy.EQ.3) then
         do 17 i=1,n
            vm(i)=0.1D0*log10((ub(i)-lb(i))/1.0D-6)
17       continue
      endif
      if(staticNeb) then
         do 18 i=1,n
            vm(i)=0.1D0*(ub(i)-lb(i))
            vm0(i)=vm(i)
18       continue
      endif

C  Determine lambda's VIM by degree of violation
      do 7 i=1,nc
         rlambda(i)=0.01D0
         vm(i+n)=rlambda(i)*sh_cst(i,n,x)
7     continue

C  Start the main loop. Note that it terminates if (i) the algorithm
C  succesfully optimizes the function or (ii) there are too many
C  function evaluations (more than MAXEVL).
      nSAIter=0

100   NUP = 0
      NREJ = 0
      NNEW = 0
      NDOWN = 0
      LNOBDS = 0

      do 110 i=1,nc
         ndecl(i)=0
         nincl(i)=0
         numnov(i) = 0
         numv(i) = 0
110   continue

      if(iaccept.EQ.3) then
         qT=2.0D0-1.0D0/exprep(T)
         qT2=1.0D0/(1.0D0-qT)
      endif


c
c   Static Penalty 
c
      if(mPen.EQ.1) then
            do j=nx+1,neq
                x(j)=1000.0D0
            enddo
            print *,"static penalty T=", T 
      endif

      DO 400, M = 1, NT

         do 350 i=1,neq
            xp(i)=x(i)
350      continue

         NTnew=0

         nSAIter=nSAIter+1

         DO 300, J = 1, NS
            DO 200, H = 1, N

C  Generate XP, the trial value of X. Note use of VM to choose XP.
C  Evaluate the function with the trial point XP and return as FP.

               if(mPen.GT.0) then
                    xp(h)=xpi(2,h,x(h),vm(h),lb(h),ub(h))
                    call fcn(n,xp,fp)
               endif 
               
               xp(h)=xpi(iscauchy,h,x(h),vm(h),lb(h),ub(h))
               call fcn(n,xp,fp)

               IF(.NOT. MAX) FP = -FP
               NFCNEV = NFCNEV + 1
               IF(IPRINT .GE. 3) CALL PRT4(MAX,NEQ,XP,X,FP,F)

C  If too many function evaluations occur, terminate the algorithm.
c               IF(NFCNEV .GE. MAXEVL) THEN
c                  CALL PRT5
c                  IF (.NOT. MAX) FOPT = -FOPT
c                  IER = 1
c                  RETURN
c               END IF

C  Accept the new point if the function value increases.
c         Metroplis or Tsallis rule
            if(iaccept.EQ.1.OR.iaccept.EQ.3) then
               IF(FP .GE. F) THEN
                  IF(IPRINT .GE. 3) THEN
                     WRITE(*,'(''  POINT ACCEPTED'')')
                  END IF
                  X(h) = XP(h)
                  F = FP
                  NACC = NACC + 1
                  NACP(H) = NACP(H) + 1
                  NUP = NUP + 1

C  If greater than any other point, record as new optimum.
                  if(sh_viomax(x).LT.vtol) then
                     fval=sh_obj(n,x)
                     if(fval.LT.fmin) then
                        IF(IPRINT .GE. 3) THEN
                           WRITE(*,'(''  NEW OPTIMUM'')')
                        END IF
                        DO 130, I = 1, neq
                           XOPT(I) = X(I)
130                     CONTINUE
                        fmin=fval
                        fopt=fp
                        NNEW = NNEW + 1
                        NTnew = NTnew + 1
c                        if(fval.LE.(-600)) return
                     end if
                  end if
               ELSE
                  if(iaccept.EQ.1) then
                     P = EXPREP((FP - F)/T)
                  else
                     P = (1.0D0+(qT-1.0D0)*(F-FP)/T)**qT2
                  endif
                  PP = RANMAR()
                  IF (PP .LT. P) THEN
                     IF(IPRINT .GE. 3) CALL PRT6(MAX)
                     X(h) = XP(h)
                     F = FP
                     NACC = NACC + 1
                     NACP(H) = NACP(H) + 1
                     NDOWN = NDOWN + 1
                  ELSE
                     xp(h)=x(h)
                     NREJ = NREJ + 1
                     IF(IPRINT .GE. 3) CALL PRT7(MAX)
                  END IF
               END IF

c         Logistic or Hastings' rule
            else 
               if(iaccept.EQ.0) then
                  P = 1.0D0/(1.0D0+EXPREP((F - FP)/T))
               else if(iaccept.EQ.2) then
                  dtf = exprep((F-FP)/T)
                  if(dtf.LT.1.0D0) then
                     P = 1.0D0+2.0D0*(0.5D0*dtf)**qH
                  else 
                     P = 1.0D0+2.0D0*(0.5D0/dtf)**qH
                  endif
                  P = P/(1.0D0+dtf)
               else
                  print *, "wrong acceptance rule in cSA"
                  stop
               endif
               PP = RANMAR()
               IF(PP .LT. P) THEN
                  X(h) = XP(h)
                  F = FP
                  NACC = NACC + 1
                  NACP(H) = NACP(H) + 1

                  if(sh_viomax(x).LT.vtol) then
                     fval=sh_obj(n,x)
                     if(fval.LT.fmin) then
                        DO 131 I = 1, neq
                           XOPT(I) = X(I)
131                     CONTINUE
                        fmin=fval
                        fopt=fp
                        NNEW = NNEW + 1
                        NTnew = NTnew + 1
                     end if
                  end if
               ELSE
                  xp(h)=x(h)
                  NREJ = NREJ + 1
               ENDIF
            endif

200         CONTINUE
300      CONTINUE

c         if(NTnew.GT.0) print *, nSAIter, "  ", fmin

C  Adjust VM so that approximately half of all evaluations are accepted.
         if(staticNeb) then
            do 311 i=1,n
               vm(i)=(vm0(i)-1.0D-6)/(t0-1.0D-6)*(t-t0)+vm0(i)
311         continue
         else
            DO 310, I = 1, N
               RATIO = DFLOAT(NACP(I))/DFLOAT(NS)
               IF (RATIO .GT. rateup) THEN
                  if(iscauchy.EQ.1) then
c                     vm(I) = vm(I)*(1. + C(I)*(RATIO - rateup)/(1.0D0-rateup))
                     vm(I) = vm(I)*(1. + 3.5D0*C(I)*(RATIO - rateup)/(1.0D0-rateup))
                  else if(iscauchy.EQ.3) then
                     vm(I)=vm(I)*2.0D0
                  else
                     vm(I) = vm(I)*(1. + C(I)*(RATIO - rateup)/(1.0D0-rateup))
                  endif
               ELSE IF (RATIO .LT. ratelow) THEN
                  if(iscauchy.EQ.3) then
                     vm(I)=vm(I)*0.5D0
                  else
                     vm(I) = vm(I)/(1. + C(I)*((ratelow - RATIO)/ratelow))
                  endif
               END IF
               IF (vm(I) .GT. (UB(I)-LB(I))) vm(I) = UB(I) - LB(I)
310         CONTINUE
         endif

         do 50 i=1,nc
            vl=sh_cst(i,n,x)
            if(vl.LT.0.01D0*t) then
                ndecl(i)=ndecl(i)+1
                nincl(i)=0
                if(ndecl(i).GE.3) then
                    ndecl(i)=0
                    rlambda(i)=rlambda(i)*0.8D0
                    if(rlambda(i).LT.0.001D0) rlambda(i)=0.001D0
                endif
            else if(vl.GT.1.0D0*t) then
                nincl(i)=nincl(i)+1
                ndecl(i)=0
                if(nincl(i).GE.3) then
                    nincl(i)=0
                    rlambda(i)=rlambda(i)*1.25D0
                endif
            else
                ndecl(i)=0
                nincl(i)=0
            endif
            vm(i+n)=rlambda(i)*vl
50       continue

         do 65 h=1,neq
            xp(h)=x(h)
65       continue


c
c  Penalty methods
c
         if(mPen.EQ.1) goto 7000
c
c  Penalty methods
c
         if(mPen.EQ.2) then
            do h=n+1,neq
                x(h)=x(h)+ 0.02 * sh_cst((h-n),n,x)
            enddo            
            goto 7000
         endif

         if(mPen.EQ.3) then
            do h=n+1,neq
                if(sh_cst((h-n),n,x).LT.vtol) then
                    numnov(h-n) = numnov(h-n) + 1
                    numv(h-n) = 0
                else   
                    numnov(h-n) = 0
                    numv(h-n) = numv(h-n) + 1
                endif
                if(numnov(h-n).GT.3) then
                    x(h) = x(h) / 1.25
                endif
                if(numv(h-n).GT.3) then
                    x(h) = x(h) * 1.5
                endif
            enddo            
            goto 7000
         endif


         do 60 h=n+1,neq
            if(sh_cst((h-n),n,x).LT.vtol) goto 60

            xp(h)=x(h)+(ranmar()*2.-1.)*vm(h)
c            xp(h)=x(h)+(ranmar()*2.-1.)*vm(h)+vm(h)/3.0D0
c            rtp=ranmar()
c            if(rtp.GT.0.33333) then
c                xp(h)=x(h)+abs((ranmar()*2.-1.))*vm(h)
c            else
c                xp(h)=x(h)-abs((ranmar()*2.-1.))*vm(h)
c            endif

            if(xp(h).LT.0.0D0) xp(h)=0.0D0
            if(xp(h).GT.1.0D20) xp(h)=1.0D20

            call fcn(n,xp,fp)
            if(.NOT.max) fp=(-fp)

c         Metroplis or Tsallis rule
            if(iaccept.EQ.1.OR.iaccept.EQ.3) then
               if(fp.LE.f) then
                  x(h)=xp(h)
                  f=fp
               else
                  if(iaccept.EQ.1) then
                      p=exprep((f-fp)/T)
                  else
                      p=(1.0D0+(qT-1.0D0)*(fp-f)/T)**qT2
                  endif
                  pp=ranmar()
                  if(pp.LT.p) then
                     x(h)=xp(h)
                     f=fp
                  else
                     xp(h)=x(h)
                  endif
               endif

c         Logistic or Hastings' rule
            else 
               if(iaccept.EQ.0) then
                  p=1.0D0/(1.0D0+exprep((fp-f)/T))
               else if(iaccept.EQ.2) then
                  dtf=exprep((fp-f)/T)
                  if(dtf.LT.1.0D0) then
                      p=1.0D0+2.0D0*(0.5D0*dtf)**qH
                  else
                      p=1.0D0+2.0D0*(0.5D0/dtf)**qH
                  endif
                  p=p/(1.0D0+dtf)
               endif
               pp=ranmar()
               if(pp.LT.p) then
                  x(h)=xp(h)
                  f=fp
               else
                  xp(h)=x(h)
               endif
            endif


60       continue

7000         IF(IPRINT .GE. 2) THEN
            CALL PRT8(NEQ,VM,XOPT,X)
         END IF

         DO 320, I = 1, N
            NACP(I) = 0
320      CONTINUE

400   CONTINUE

      IF(IPRINT .GE. 1) THEN
         CALL PRT9(MAX,NEQ,T,XOPT,VM,FOPT,NUP,NDOWN,NREJ,LNOBDS,NNEW)
      END IF

      do 55 i=1,n
         vm(i)=Min(1.0D0,(ub(i)-lb(i)))
55    continue

C  Check termination criteria.
      QUIT = .FALSE.
      FSTAR(1) = F
      IF ((FOPT - FSTAR(1)) .LE. EPS) QUIT = .TRUE.
      DO 410, I = 1, NEPS
         IF (ABS(F - FSTAR(I)) .GT. EPS) QUIT = .FALSE.
410   CONTINUE

      if(t.LT.1.0D-6.OR.quit) then
         xvio=sh_viomax(x)
         if(xvio.GT.vtol.AND.nreSA.GT.0) then
            nreSA=nreSA-1
            quit=.FALSE.
            t=xvio*100.0D0
            if(t.GT.t0) t=t0
            print *, "Re-annealing from T =", t
            goto 100
         else
            quit=.TRUE.
         endif
      endif

C  Terminate SA if appropriate.
      IF (QUIT) THEN
c         DO 420, I = 1, N
c            X(I) = XOPT(I)
c420      CONTINUE
         IER = 0
         IF (.NOT. MAX) FOPT = -FOPT
         IF(IPRINT .GE. 1) CALL PRT10
         RETURN
      END IF

C  If termination criteria is not met, prepare for another loop.
      T = RT*T
      DO 430, I = NEPS, 2, -1
         FSTAR(I) = FSTAR(I-1)
430   CONTINUE

      if(nnew.GE.1) then
         F = FOPT
         DO 440, I = 1, N
            X(I) = XOPT(I)
440      CONTINUE
      endif

C  Loop again.
      GO TO 100

      END

